"""Multi apps example with more complex inline configuration."""

from extrude.multi_app import dispatch_child_apps_from_module

import example_apps_complex

config = {
    'from_func_list': {'display_name': 'From_a_list_of_functions',},
}

dispatch_child_apps_from_module(example_apps_complex, config)



"""Entrypoint for dispatching multiple child applications with a navigation root.

How to prepare an application for dispatch:

"""

import os
from functools import partial
from pathlib import Path
import streamlit as st
from typing import Iterable
import importlib.util
import sys
import types

from streamlitfront.base import dispatch_funcs

ROOT_APP = '__extrude_root__'
EXTRUDE_FUNCS = 'extrude_funcs'


def execute_module_spec(spec):
    """Imports a module into the current environment from a module spec and returns the module"""
    module = importlib.util.module_from_spec(spec)
    sys.modules[module.__name__] = module
    spec.loader.exec_module(module)
    return module


def dispatch_raw_module(spec):
    """Creates a callable that will execute a module when called.
    Used to run a streamlit app without needing to modify it.

    :param spec: importlib.machinery.ModuleSpec
    """

    def app():
        return execute_module_spec(spec)

    return app


def get_module_spec_from_pathname(pathname: str):
    """Creates an instance of importlib.machinery.ModuleSpec from a path."""
    module_name = os.path.basename(os.path.normpath(pathname))
    if os.path.isdir(pathname):
        pathname = os.path.join(pathname, '__init__.py')
    return importlib.util.spec_from_file_location(module_name, pathname)


def set_current_module(app_name):
    st.session_state['current_app'] = app_name


def render_root_nav(app_name_mapping: dict):
    st.set_page_config(layout='centered')
    st.header('Choose an app')
    for app_name, display_name in app_name_mapping.items():
        st.button(
            key=app_name,
            label=display_name,
            on_click=partial(set_current_module, app_name),
        )


def mk_dflt_dispatch(module, configs):
    funcs = getattr(module, configs.get('funcs_to_dispatch', EXTRUDE_FUNCS), [])

    def dispatch():
        return dispatch_funcs(funcs, configs=configs)

    return dispatch


def dispatch_child_apps(pathnames: Iterable[str], configs: dict = None):
    """Recurses through a Python module to collect objects that can be dispatched to a streamlit app.

    :param pathnames: An iterable of files and/or directory to scan through. If a filename
        points to a directory (a package), will look for __init__.py within that directory.
    :param configs: (Optional) A dict that maps package names with configuration settings,
        allowing you to specify which object to import from the package and what function
        to use to dispatch that object to streamlit.

    Example config:
        {"child_pkg_1": {"import_name": "funcs_to_dispatch", "dispatcher": my_dispatcher}}
        config[pkgname]['dispatch'] can be a string to refer to a dispatcher within the module,
        or a reference to a callable from the root app. The default dispatcher will be
        streamlitfront.base.dispatch_funcs.
    """
    if not configs:
        configs = {}

    module_specs = [get_module_spec_from_pathname(pathname) for pathname in pathnames]

    def get_display_name(spec):
        module_config = configs.get(spec.name, configs)
        return module_config.get('display_name', spec.name)

    app_name_mapping = {spec.name: get_display_name(spec) for spec in module_specs}
    module_mapping = {spec.name: spec for spec in module_specs}
    if 'current_app' not in st.session_state:
        st.session_state['current_app'] = ROOT_APP

    current_app_name = st.session_state['current_app']
    if current_app_name == ROOT_APP:
        return render_root_nav(app_name_mapping)

    current_module_spec = module_mapping[current_app_name]
    config_for_module = configs.get(current_app_name, configs)
    app = config_for_module.get('app', None)
    if not app:
        app = dispatch_raw_module(current_module_spec)
    current_module = execute_module_spec(current_module_spec)
    if isinstance(app, str):
        app = getattr(current_module, app, config_for_module.get(app, None))
    if not callable(app):
        dispatch = config_for_module.get(
            'dispatch', mk_dflt_dispatch(current_module, config_for_module)
        )
        if isinstance(dispatch, str):
            dispatch = getattr(current_module, dispatch)
        app = dispatch()
    try:
        app()
    except Exception as error:
        st.error(str(error))
    st.sidebar.button(
        label='Back to root',
        key='backtoroot',
        on_click=lambda: set_current_module(ROOT_APP),
    )


def dispatch_child_apps_from_root(root_dir: str, configs: dict = None):
    """Lists the contents of a directory and passes the list to dispatch_child_apps_from_paths.

    :param root_dir: A path to a directory that contains one or more children to dispatch.
    :param configs: (Optional) See dispatch_child_apps.
    """
    root_pathname = os.path.abspath(root_dir)
    if not os.path.isdir(root_pathname):
        raise ValueError(f'{root_dir} is not a path to a directory.')
    children = [path for path in os.listdir(root_pathname) if not path.startswith('__')]
    child_paths = [os.path.join(root_pathname, child) for child in children]
    return dispatch_child_apps(child_paths, configs)


def dispatch_child_apps_from_module(
    root_module: types.ModuleType, configs: dict = None
):
    """Takes a root Python module and scans through its immediate children to dispatch functions.

    :param root_module: A module that contains one or more child packages to dispatch.
    :param configs: (Optional) See dispatch_child_apps.

    >>> import streamlit as st
    >>> import extrude.examples.example_apps_simple as example_apps
    >>>
    >>> # this function call won't work correctly in the test environment
    >>> # dispatch_child_apps_from_module(example_apps)
    """
    root_filename = root_module.__file__
    root_dir = Path(root_filename).parent.absolute()
    return dispatch_child_apps_from_root(
        str(root_dir).replace('__pycache__/', ''), configs
    )



import json
from functools import wraps
from typing import Callable, Iterable, Union

import streamlit.bootstrap
from http2py import HttpClient
from i2 import name_of_obj
from i2.wrapper import Sig
from py2http import mk_app as mk_webservice
from py2http import run_app as run_webservice
from py2http.util import run_process
from streamlitfront.base import dispatch_funcs
from meshed import DAG


# TODO: Maybe split this function in two, by adding mk_app_from_openapi_spec
def mk_app(
    obj: Union[DAG, Iterable[Callable]] = None,
    *,
    api_url: str = None,
    openapi_spec: dict = None,
    func_names: Iterable[str] = None,
    **kwargs
):
    """Generates a front application which will consume a web service exposing a bunch
    of functions. There are three ways to call this function: from a dag, a list of
    functions or directly from the openapi specification of the web service and a list
    of function names to pick from the openapi specification.

    :param obj: A dag or list of functions. Overwrites ``openapi_spec`` and
    ``func_names`` if not None.
    :param api_url: The base url of the API. Only used if ``obj`` is not None.
    :param openapi_spec: A web service openapi specification. Ignored if ``obj``
    is not None.
    :param func_names: A list of function names to pick from the openapi specification.
    Ignored if ``obj`` is not None.
    :param kwargs: Any extra keyword argument used to make the front application.

    >>> def foo():
    ...     pass
    ...
    >>> app = mk_app([foo])
    >>>
    >>> ws = mk_webservice([foo])
    >>> app = mk_app(
    ...     openapi_spec=ws.openapi_spec,
    ...     func_names=['foo']
    ... )
    """

    def _flatten(meth):
        sig = Sig(meth) - 'self'

        @sig
        @wraps(meth)
        def flat_func(*args, **kwargs):
            return meth(*args, **kwargs)

        return flat_func

    if obj:
        ws_config = dict(openapi=dict(base_url=api_url)) if api_url else {}
        funcs = [fn.func for fn in obj.func_nodes] if isinstance(obj, DAG) else obj
        ws = mk_webservice(funcs, **ws_config)
        openapi_spec = ws.openapi_spec
        func_names = [name_of_obj(func) for func in funcs]
    api = HttpClient(openapi_spec=openapi_spec)
    ws_funcs = [_flatten(getattr(api, name)) for name in func_names]

    return dispatch_funcs(ws_funcs, **kwargs)


def run_app(obj: Union[DAG, Iterable[Callable]], *, ws_config: dict = None, **kwargs):
    """Runs a extrude application from a bunch of fucntions. It first run a web service
    application from those functions then runs a front applciation to consume these web
    services

    :param obj: A dag or a list of functions.
    :param ws_config: A predefined configuration for the web service application.
    :param kwargs: Any extra keyword argument used to make the front application.
    """

    ws_config = ws_config or {}
    funcs = [fn.func for fn in obj.func_nodes] if isinstance(obj, DAG) else obj

    with run_process(
        func=run_webservice, func_kwargs=dict(app_obj=funcs, **ws_config), is_ready=3,
    ):
        ws = mk_webservice(funcs)
        func_names = [name_of_obj(func) for func in funcs]
        streamlit.bootstrap.run(
            __file__,
            args=[
                json.dumps(ws.openapi_spec),
                json.dumps(func_names),
                json.dumps(kwargs),
            ],
            command_line='',
            flag_options={},
        )


if __name__ == '__main__':
    from sys import argv

    openapi_spec = json.loads(argv[1])
    func_names = json.loads(argv[2])
    kwargs = json.loads(argv[3])

    app = mk_app(openapi_spec=openapi_spec, func_names=func_names, **kwargs,)
    app()



from extrude.run_app import run_app


def foo(a: int = 0, b: int = 0, c=0):
    """This is foo. It computes something"""
    return (a * b) + c


def bar(x, greeting='hello'):
    """bar greets its input"""
    return f'{greeting} {x}'


def confuser(a: int = 0, x: float = 3.14):
    return (a ** 2) * x


funcs = [foo, bar, confuser]


if __name__ == '__main__':
    run_app(funcs)



"""
Control complex python object through strings.
Wrap functions so that the complex arguments can be specified through a string key
that points to the actual python object (which is stored in a session's memory or
persisted in some fashion).
"""

from warnings import warn

warn(f'Module has moved to front.crude')

from front.crude import *



"""
Little streamlit app to help make decision on whether to invest in real estate vs stock market
"""

import streamlit as st
import pandas as pd
from investate.real_estate_vs_stock import (
    house_investment,
    compare_house_invest_vs_stock,
    compute_mortg_principal,
)

st.title('House vs other investment')

house_cost = st.sidebar.number_input('House price', value=240000)
down_payment_perc = st.sidebar.number_input('Down payment percentage', value=10) * 0.01
mortg_rate = st.sidebar.number_input('Mortgage rate', value=2.75) * 0.01
mortgage_n_years = st.sidebar.number_input('Mortgage duration in years', value=15)
n_years_after_pay_off = st.sidebar.number_input(
    'Years to display after pay off', value=5
)

tax = st.sidebar.number_input('Yearly tax', value=4000)
insurance = st.sidebar.number_input('Yearly insurance cost', value=3000)
repair = st.sidebar.number_input('Yearly repair cost', value=6000)
monthly_rental_income = st.sidebar.number_input(
    'Average monthly rental income', value=7500
)
percentage_rented = st.sidebar.number_input('Percentage rented', value=0.5)

estate_rate = (
    st.sidebar.number_input('Yearly real estate market increase', value=3.5) * 0.01
)
inflation_rate = st.sidebar.number_input('Inflation rate', value=2.1) * 0.01
income_tax = st.sidebar.number_input('Income tax percentage', value=33) * 0.01
management_fees_rate = st.sidebar.number_input('Management fees', value=22) * 0.01
stock_market_rate = (
    st.sidebar.number_input('Yearly rate of other investment', value=8) * 0.01
)

equity, monthly_income = house_investment(
    mortg_rate,
    down_payment_perc,
    house_cost,
    tax,
    insurance,
    repair,
    estate_rate,
    mortgage_n_years,
    n_years_after_pay_off,
    monthly_rental_income,
    percentage_rented,
    inflation_rate,
    income_tax,
    management_fees_rate,
    plot=False,
)

df = pd.DataFrame()
df['equity'] = equity
df['montlhy_income'] = monthly_income

house_invest, down_payment_invest = compare_house_invest_vs_stock(
    equity,
    monthly_income,
    stock_market_rate=stock_market_rate,
    down_payment_perc=down_payment_perc,
    house_cost=house_cost,
    plot=False,
)
df['house_investment'] = house_invest
df['stock_investment'] = down_payment_invest

st.line_chart(df['montlhy_income'])
st.line_chart(df['equity'])
st.line_chart(df[['house_investment', 'stock_investment']])

# monthly_mort_payment = compute_mortg_principal(loan_rate=mortg_rate,
#                                                loan_amount=house_cost - down_payment_perc * house_cost,
#                                                years_to_maturity=mortgage_n_years,
#                                                n_payment_per_year=12, )
# # Figure size
# fig = plt.figure(figsize=(5, 5))
# ax = fig.add_subplot(111)
# ax.plot(df['montlhy_income'], label='montlhy income')
# ax.vlines(
#     x=mortgage_n_years,
#     ymin=np.min(equity),
#     ymax=np.max(equity),
#     label='mortgage paid off',
#     linestyles='dashed',
#     linewidth=0.5,
# )
# ax.set_xlabel('months')
# ax.set_ylabel('total')
# ax.set_title(
#     f'Equity over {mortgage_n_years + n_years_after_pay_off} years'
# )
# ax.legend()
#
# ax.plot(monthly_income, label='monthly income')
# ax.vlines(
#     x=mortgage_n_years,
#     ymin=np.min(monthly_income + [monthly_mort_payment]),
#     ymax=np.max(monthly_income + [monthly_mort_payment]),
#     label='mortgage paid off',
#     linestyles='dashed',
#     linewidth=0.5,
# )
# ax.hlines(
#     y=monthly_mort_payment,
#     xmin=0,
#     xmax=n_years_after_pay_off + n_years_after_pay_off,
#     label='monthly mortgage payment',
#     linestyles='dashed',
#     linewidth=0.5,
#     color='r',
# )
# ax.set_xlabel('months')
# ax.set_ylabel('monthly income')
# ax.legend()
# ax.set_title(
#     f'Monthly income over {mortgage_n_years + n_years_after_pay_off} years'
# )
# st.pyplot(fig, dpi=500)
